@@ -28,12 +28,51 @@ import textwrap
 from .. import mlog, mesonlib
 from ..mesonlib import PerMachine, Popen_safe, version_compare, MachineChoice
 from ..environment import Environment
+from ..envconfig import get_env_var
+from ..compilers import (
+    AppleClangCCompiler, AppleClangCPPCompiler, AppleClangObjCCompiler,
+    AppleClangObjCPPCompiler
+)
 
 if T.TYPE_CHECKING:
     from ..dependencies.base import ExternalProgram
+    from ..compilers import Compiler
 
 TYPE_result = T.Tuple[int, T.Optional[str], T.Optional[str]]
 
+_MESON_TO_CMAKE_MAPPING = {
+    'arm':          'ARMCC',
+    'armclang':     'ARMClang',
+    'clang':        'Clang',
+    'clang-cl':     'MSVC',
+    'flang':        'Flang',
+    'g95':          'G95',
+    'gcc':          'GNU',
+    'intel':        'Intel',
+    'intel-cl':     'MSVC',
+    'msvc':         'MSVC',
+    'pathscale':    'PathScale',
+    'pgi':          'PGI',
+    'sun':          'SunPro',
+}
+
+def meson_compiler_to_cmake_id(cobj: 'Compiler') -> str:
+    """Translate meson compiler's into CMAKE compiler ID's.
+
+    Most of these can be handled by a simple table lookup, with a few
+    exceptions.
+
+    Clang and Apple's Clang are both identified as "clang" by meson. To make
+    things more complicated gcc and vanilla clang both use Apple's ld64 on
+    macOS. The only way to know for sure is to do an isinstance() check.
+    """
+    if isinstance(cobj, (AppleClangCCompiler, AppleClangCPPCompiler,
+                         AppleClangObjCCompiler, AppleClangObjCPPCompiler)):
+        return 'AppleClang'
+    # If no mapping, try GNU and hope that the build files don't care
+    return _MESON_TO_CMAKE_MAPPING.get(cobj.get_id(), 'GNU')
+
+
 class CMakeExecutor:
     # The class's copy of the CMake path. Avoids having to search for it
     # multiple times in the same Meson invocation.
@@ -48,6 +87,8 @@ class CMakeExecutor:
         self.cmakebin, self.cmakevers = self.find_cmake_binary(self.environment, silent=silent)
         self.always_capture_stderr = True
         self.print_cmout = False
+        self.prefix_paths = []      # type: T.List[str]
+        self.extra_cmake_args = []  # type: T.List[str]
         if self.cmakebin is False:
             self.cmakebin = None
             return
@@ -60,26 +101,28 @@ class CMakeExecutor:
             self.cmakebin = None
             return
 
-    def find_cmake_binary(self, environment: Environment, silent: bool = False) -> T.Tuple['ExternalProgram', str]:
-        from ..dependencies.base import ExternalProgram
+        self.prefix_paths = self.environment.coredata.builtins_per_machine[self.for_machine]['cmake_prefix_path'].value
+        env_pref_path = get_env_var(
+            self.for_machine,
+            self.environment.is_cross_build(),
+            'CMAKE_PREFIX_PATH')
+        if env_pref_path is not None:
+            if mesonlib.is_windows():
+                # Cannot split on ':' on Windows because its in the drive letter
+                env_pref_path = env_pref_path.split(os.pathsep)
+            else:
+                # https://github.com/mesonbuild/meson/issues/7294
+                env_pref_path = re.split(r':|;', env_pref_path)
+            env_pref_path = [x for x in env_pref_path if x]  # Filter out empty strings
+            if not self.prefix_paths:
+                self.prefix_paths = []
+            self.prefix_paths += env_pref_path
 
-        # Create an iterator of options
-        def search():
-            # Lookup in cross or machine file.
-            potential_cmakepath = environment.lookup_binary_entry(self.for_machine, 'cmake')
-            if potential_cmakepath is not None:
-                mlog.debug('CMake binary for %s specified from cross file, native file, or env var as %s.', self.for_machine, potential_cmakepath)
-                yield ExternalProgram.from_entry('cmake', potential_cmakepath)
-                # We never fallback if the user-specified option is no good, so
-                # stop returning options.
-                return
-            mlog.debug('CMake binary missing from cross or native file, or env var undefined.')
-            # Fallback on hard-coded defaults.
-            # TODO prefix this for the cross case instead of ignoring thing.
-            if environment.machines.matches_build_machine(self.for_machine):
-                for potential_cmakepath in environment.default_cmake:
-                    mlog.debug('Trying a default CMake fallback at', potential_cmakepath)
-                    yield ExternalProgram(potential_cmakepath, silent=True)
+        if self.prefix_paths:
+            self.extra_cmake_args += ['-DCMAKE_PREFIX_PATH={}'.format(';'.join(self.prefix_paths))]
+
+    def find_cmake_binary(self, environment: Environment, silent: bool = False) -> T.Tuple['ExternalProgram', str]:
+        from ..dependencies.base import find_external_program
 
         # Only search for CMake the first time and store the result in the class
         # definition
@@ -89,10 +132,11 @@ class CMakeExecutor:
             mlog.debug('CMake binary for %s is cached.' % self.for_machine)
         else:
             assert CMakeExecutor.class_cmakebin[self.for_machine] is None
+
             mlog.debug('CMake binary for %s is not cached' % self.for_machine)
-            for potential_cmakebin in search():
-                mlog.debug('Trying CMake binary {} for machine {} at {}'
-                           .format(potential_cmakebin.name, self.for_machine, potential_cmakebin.command))
+            for potential_cmakebin in find_external_program(
+                    environment, self.for_machine, 'cmake', 'CMake',
+                    environment.default_cmake, allow_default_for_cross=False):
                 version_if_ok = self.check_cmake(potential_cmakebin)
                 if not version_if_ok:
                     continue
@@ -132,7 +176,7 @@ class CMakeExecutor:
                 msg += '\n\nOn Unix-like systems this is often caused by scripts that are not executable.'
             mlog.warning(msg)
             return None
-        cmvers = re.sub(r'\s*(cmake|cmake3) version\s*', '', out.split('\n')[0]).strip()
+        cmvers = re.search(r'(cmake|cmake3)\s*version\s*([\d.]+)', out).group(2)
         return cmvers
 
     def set_exec_mode(self, print_cmout: T.Optional[bool] = None, always_capture_stderr: T.Optional[bool] = None) -> None:
@@ -226,6 +270,7 @@ class CMakeExecutor:
         if env is None:
             env = os.environ
 
+        args = args + self.extra_cmake_args
         if disable_cache:
             return self._call_impl(args, build_dir, env)
 
@@ -260,29 +305,33 @@ class CMakeExecutor:
                 p = fallback
             return p
 
-        def choose_compiler(lang: str) -> T.Tuple[str, str]:
+        def choose_compiler(lang: str) -> T.Tuple[str, str, str, str]:
+            comp_obj = None
             exe_list = []
             if lang in compilers:
-                exe_list = compilers[lang].get_exelist()
+                comp_obj = compilers[lang]
             else:
                 try:
                     comp_obj = self.environment.compiler_from_language(lang, MachineChoice.BUILD)
-                    if comp_obj is not None:
-                        exe_list = comp_obj.get_exelist()
                 except Exception:
                     pass
 
+            if comp_obj is not None:
+                exe_list = comp_obj.get_exelist()
+                comp_id = meson_compiler_to_cmake_id(comp_obj)
+                comp_version = comp_obj.version.upper()
+
             if len(exe_list) == 1:
-                return make_abs(exe_list[0], lang), ''
+                return make_abs(exe_list[0], lang), '', comp_id, comp_version
             elif len(exe_list) == 2:
-                return make_abs(exe_list[1], lang), make_abs(exe_list[0], lang)
+                return make_abs(exe_list[1], lang), make_abs(exe_list[0], lang), comp_id, comp_version
             else:
                 mlog.debug('Failed to find a {} compiler for CMake. This might cause CMake to fail.'.format(lang))
-                return fallback, ''
+                return fallback, '', 'GNU', ''
 
-        c_comp, c_launcher = choose_compiler('c')
-        cxx_comp, cxx_launcher = choose_compiler('cpp')
-        fortran_comp, fortran_launcher = choose_compiler('fortran')
+        c_comp, c_launcher, c_id, c_version = choose_compiler('c')
+        cxx_comp, cxx_launcher, cxx_id, cxx_version = choose_compiler('cpp')
+        fortran_comp, fortran_launcher, _, _ = choose_compiler('fortran')
 
         # on Windows, choose_compiler returns path with \ as separator - replace by / before writing to CMAKE file
         c_comp = c_comp.replace('\\', '/')
@@ -304,34 +353,42 @@ class CMakeExecutor:
         fortran_comp_file = comp_dir / 'CMakeFortranCompiler.cmake'
 
         if c_comp and not c_comp_file.is_file():
+            is_gnu = '1' if c_id == 'GNU' else ''
             c_comp_file.write_text(textwrap.dedent('''\
                 # Fake CMake file to skip the boring and slow stuff
                 set(CMAKE_C_COMPILER "{}") # Should be a valid compiler for try_compile, etc.
                 set(CMAKE_C_COMPILER_LAUNCHER "{}") # The compiler launcher (if presentt)
-                set(CMAKE_C_COMPILER_ID "GNU") # Pretend we have found GCC
-                set(CMAKE_COMPILER_IS_GNUCC 1)
+                set(CMAKE_COMPILER_IS_GNUCC {})
+                set(CMAKE_C_COMPILER_ID "{}")
+                set(CMAKE_C_COMPILER_VERSION "{}")
                 set(CMAKE_C_COMPILER_LOADED 1)
+                set(CMAKE_C_COMPILER_FORCED 1)
                 set(CMAKE_C_COMPILER_WORKS TRUE)
                 set(CMAKE_C_ABI_COMPILED TRUE)
                 set(CMAKE_C_SOURCE_FILE_EXTENSIONS c;m)
                 set(CMAKE_C_IGNORE_EXTENSIONS h;H;o;O;obj;OBJ;def;DEF;rc;RC)
                 set(CMAKE_SIZEOF_VOID_P "{}")
-            '''.format(c_comp, c_launcher, ctypes.sizeof(ctypes.c_voidp))))
+            '''.format(c_comp, c_launcher, is_gnu, c_id, c_version,
+                       ctypes.sizeof(ctypes.c_voidp))))
 
         if cxx_comp and not cxx_comp_file.is_file():
+            is_gnu = '1' if cxx_id == 'GNU' else ''
             cxx_comp_file.write_text(textwrap.dedent('''\
                 # Fake CMake file to skip the boring and slow stuff
                 set(CMAKE_CXX_COMPILER "{}") # Should be a valid compiler for try_compile, etc.
                 set(CMAKE_CXX_COMPILER_LAUNCHER "{}") # The compiler launcher (if presentt)
-                set(CMAKE_CXX_COMPILER_ID "GNU") # Pretend we have found GCC
-                set(CMAKE_COMPILER_IS_GNUCXX 1)
+                set(CMAKE_COMPILER_IS_GNUCXX {})
+                set(CMAKE_CXX_COMPILER_ID "{}")
+                set(CMAKE_CXX_COMPILER_VERSION "{}")
                 set(CMAKE_CXX_COMPILER_LOADED 1)
+                set(CMAKE_CXX_COMPILER_FORCED 1)
                 set(CMAKE_CXX_COMPILER_WORKS TRUE)
                 set(CMAKE_CXX_ABI_COMPILED TRUE)
                 set(CMAKE_CXX_IGNORE_EXTENSIONS inl;h;hpp;HPP;H;o;O;obj;OBJ;def;DEF;rc;RC)
                 set(CMAKE_CXX_SOURCE_FILE_EXTENSIONS C;M;c++;cc;cpp;cxx;mm;CPP)
                 set(CMAKE_SIZEOF_VOID_P "{}")
-            '''.format(cxx_comp, cxx_launcher, ctypes.sizeof(ctypes.c_voidp))))
+            '''.format(cxx_comp, cxx_launcher, is_gnu, cxx_id, cxx_version,
+                       ctypes.sizeof(ctypes.c_voidp))))
 
         if fortran_comp and not fortran_comp_file.is_file():
             fortran_comp_file.write_text(textwrap.dedent('''\
@@ -362,5 +419,8 @@ class CMakeExecutor:
     def get_command(self) -> T.List[str]:
         return self.cmakebin.get_command()
 
+    def get_cmake_prefix_paths(self) -> T.List[str]:
+        return self.prefix_paths
+
     def machine_choice(self) -> MachineChoice:
         return self.for_machine
