@@ -19,13 +19,13 @@ from ..mesonlib import (
     EnvironmentException, MachineChoice, version_compare,
 )
 
+from ..arglist import CompilerArgs
 from .compilers import (
     d_dmd_buildtype_args,
     d_gdc_buildtype_args,
     d_ldc_buildtype_args,
     clike_debug_args,
     Compiler,
-    CompilerArgs,
 )
 from .mixins.gnu import GnuCompiler
 
@@ -105,6 +105,9 @@ class DmdLikeCompilerMixin:
     def get_coverage_args(self):
         return ['-cov']
 
+    def get_coverage_link_args(self):
+        return []
+
     def get_preprocess_only_args(self):
         return ['-E']
 
@@ -217,7 +220,7 @@ class DmdLikeCompilerMixin:
 
     def build_rpath_args(self, env, build_dir, from_dir, rpath_paths, build_rpath, install_rpath):
         if self.info.is_windows():
-            return []
+            return ([], set())
 
         # GNU ld, solaris ld, and lld acting like GNU ld
         if self.linker.id.startswith('ld'):
@@ -225,15 +228,16 @@ class DmdLikeCompilerMixin:
             # do directly, each argument -rpath and the value to rpath, need to be
             # split into two separate arguments both prefaced with the -L=.
             args = []
-            for r in super().build_rpath_args(
-                    env, build_dir, from_dir, rpath_paths, build_rpath, install_rpath):
+            (rpath_args, rpath_dirs_to_remove) = super().build_rpath_args(
+                    env, build_dir, from_dir, rpath_paths, build_rpath, install_rpath)
+            for r in rpath_args:
                 if ',' in r:
                     a, b = r.split(',', maxsplit=1)
                     args.append(a)
                     args.append(self.LINKER_PREFIX + b)
                 else:
                     args.append(r)
-            return args
+            return (args, rpath_dirs_to_remove)
 
         return super().build_rpath_args(
             env, build_dir, from_dir, rpath_paths, build_rpath, install_rpath)
@@ -423,6 +427,9 @@ class DmdLikeCompilerMixin:
             args = [a.replace('-L=', '-Xcc=-Wl,') for a in args]
         return args
 
+class DCompilerArgs(CompilerArgs):
+    prepend_prefixes = ('-I', '-L')
+    dedup2_prefixes = ('-I')
 
 class DCompiler(Compiler):
     mscrt_args = {
@@ -578,7 +585,7 @@ class DCompiler(Compiler):
         elif not isinstance(dependencies, list):
             dependencies = [dependencies]
         # Collect compiler arguments
-        args = CompilerArgs(self)
+        args = self.compiler_args()
         for d in dependencies:
             # Add compile flags needed by dependencies
             args += d.get_compile_args()
@@ -596,6 +603,9 @@ class DCompiler(Compiler):
         args += extra_args
         return args
 
+    def compiler_args(self, args: T.Optional[T.Iterable[str]] = None) -> DCompilerArgs:
+        return DCompilerArgs(self, args)
+
     def compiles(self, code, env, *, extra_args=None, dependencies=None, mode='compile'):
         args = self._get_compiler_check_args(env, extra_args, dependencies, mode)
 
@@ -642,7 +652,8 @@ class GnuDCompiler(GnuCompiler, DCompiler):
                           '1': default_warn_args,
                           '2': default_warn_args + ['-Wextra'],
                           '3': default_warn_args + ['-Wextra', '-Wpedantic']}
-        self.base_options = ['b_colorout', 'b_sanitize', 'b_staticpic', 'b_vscrt']
+        self.base_options = ['b_colorout', 'b_sanitize', 'b_staticpic',
+                             'b_vscrt', 'b_coverage', 'b_pgo', 'b_ndebug']
 
         self._has_color_support = version_compare(self.version, '>=4.9')
         # dependencies were implemented before, but broken - support was fixed in GCC 7.1+
@@ -662,9 +673,6 @@ class GnuDCompiler(GnuCompiler, DCompiler):
     def get_warn_args(self, level):
         return self.warn_args[level]
 
-    def get_coverage_args(self):
-        return []
-
     def get_buildtype_args(self, buildtype):
         return d_gdc_buildtype_args[buildtype]
 
@@ -684,6 +692,9 @@ class GnuDCompiler(GnuCompiler, DCompiler):
             return args
         return args + ['-shared-libphobos']
 
+    def get_disable_assert_args(self):
+        return ['-frelease']
+
 
 class LLVMDCompiler(DmdLikeCompilerMixin, DCompiler):
 
@@ -691,7 +702,7 @@ class LLVMDCompiler(DmdLikeCompilerMixin, DCompiler):
                  info: 'MachineInfo', arch, **kwargs):
         DCompiler.__init__(self, exelist, version, for_machine, info, arch, False, None, **kwargs)
         self.id = 'llvm'
-        self.base_options = ['b_coverage', 'b_colorout', 'b_vscrt']
+        self.base_options = ['b_coverage', 'b_colorout', 'b_vscrt', 'b_ndebug']
 
     def get_colorout_args(self, colortype):
         if colortype == 'always':
@@ -733,6 +744,9 @@ class LLVMDCompiler(DmdLikeCompilerMixin, DCompiler):
             return args
         return args + ['-link-defaultlib-shared']
 
+    def get_disable_assert_args(self) -> T.List[str]:
+        return ['--release']
+
 
 class DmdDCompiler(DmdLikeCompilerMixin, DCompiler):
 
@@ -740,7 +754,7 @@ class DmdDCompiler(DmdLikeCompilerMixin, DCompiler):
                  info: 'MachineInfo', arch, **kwargs):
         DCompiler.__init__(self, exelist, version, for_machine, info, arch, False, None, **kwargs)
         self.id = 'dmd'
-        self.base_options = ['b_coverage', 'b_colorout', 'b_vscrt']
+        self.base_options = ['b_coverage', 'b_colorout', 'b_vscrt', 'b_ndebug']
 
     def get_colorout_args(self, colortype):
         if colortype == 'always':
@@ -803,3 +817,6 @@ class DmdDCompiler(DmdLikeCompilerMixin, DCompiler):
         if self.info.is_windows():
             return args
         return args + ['-defaultlib=phobos2', '-debuglib=phobos2']
+
+    def get_disable_assert_args(self) -> T.List[str]:
+        return ['-release']
