@@ -53,7 +53,7 @@ def add_arguments(parser):
                              'is not working.')
     parser.add_argument('--wipe', action='store_true',
                         help='Wipe build directory and reconfigure using previous command line options. ' +
-                             'Userful when build directory got corrupted, or when rebuilding with a ' +
+                             'Useful when build directory got corrupted, or when rebuilding with a ' +
                              'newer version of meson.')
     parser.add_argument('builddir', nargs='?', default=None)
     parser.add_argument('sourcedir', nargs='?', default=None)
@@ -86,7 +86,7 @@ class MesonApp:
                     # will cause a crash
                     for l in os.listdir(self.build_dir):
                         l = os.path.join(self.build_dir, l)
-                        if os.path.isdir(l):
+                        if os.path.isdir(l) and not os.path.islink(l):
                             mesonlib.windows_proof_rmtree(l)
                         else:
                             mesonlib.windows_proof_rm(l)
